@@ -13,32 +13,26 @@ require DynaLoader;
 	write_png
 	init_io
 	read_info
+	read_update_info
+	read_image
 	read_png
-	get_IHDR
-	get_tIME
-	set_tIME
 	get_text
 	set_text
 	sig_cmp
+	scalar_as_input
 	read_from_scalar
 	color_type_name
 	text_compression_name
 	get_libpng_ver
 	access_version_number
+	get_row_pointers
 	get_rows
 	get_rowbytes
-	get_PLTE
-	set_PLTE
-	get_bKGD
-	get_cHRM
 	get_channels
-	get_sBIT
-	get_oFFs
-	get_pHYs
-	get_sRGB
 	get_valid
+	set_row_pointers
+	set_tRNS_pointer
 	set_rows
-	set_IHDR
 	write_to_scalar
 	set_filter
 	set_verbosity
@@ -46,6 +40,43 @@ require DynaLoader;
 	get_unknown_chunks
 	supports
 	set_keep_unknown_chunks
+	get_tRNS_palette
+	set_PLTE_pointer
+	set_expand
+	set_gray_to_rgb
+	set_filler
+	get_sRGB
+	set_sRGB
+	set_packing
+	set_strip_16
+	get_bKGD
+	set_bKGD
+	get_cHRM
+	set_cHRM
+	get_gAMA
+	set_gAMA
+	get_iCCP
+	set_iCCP
+	get_IHDR
+	set_IHDR
+	get_oFFs
+	set_oFFs
+	get_pCAL
+	set_pCAL
+	get_pHYs
+	set_pHYs
+	get_PLTE
+	set_PLTE
+	get_sBIT
+	set_sBIT
+	get_sCAL
+	set_sCAL
+	get_sPLT
+	set_sPLT
+	get_tIME
+	set_tIME
+	get_tRNS
+	set_tRNS
 /;
 
 %EXPORT_TAGS = (
@@ -53,7 +84,7 @@ require DynaLoader;
 );
 
 require XSLoader;
-our $VERSION = 0.03;
+our $VERSION = 0.04;
 
 XSLoader::load('Image::PNG', $VERSION);
 
@@ -210,7 +241,7 @@ a routine which reads data from Perl scalars. It then uses
 C<png_read_png> to read all the data.
 
 The C function which does this is called C<perl_png_scalar_read>, 
-L<in the file C<perl-libpng.c> in the top directory of the distribution|http://cpansearch.perl.org/src/BKB/Image-PNG-0.03/perl-libpng.c>.
+L<in the file C<perl-libpng.c> in the top directory of the distribution|http://cpansearch.perl.org/src/BKB/Image-PNG-0.04/perl-libpng.c>.
 
 See also L</Input/output manipulation functions>.
 
@@ -240,7 +271,7 @@ uses C<png_set_write_fn> to set the writing function of C<$png> to be
 its own function, which writes data to the Perl scalar.
 
 The C function which does this is called C<perl_png_scalar_write>, 
-L<in the file C<perl-libpng.c> in the top directory of the distribution|http://cpansearch.perl.org/src/BKB/Image-PNG-0.03/perl-libpng.c>.
+L<in the file C<perl-libpng.c> in the top directory of the distribution|http://cpansearch.perl.org/src/BKB/Image-PNG-0.04/perl-libpng.c>.
 
 See also L</Input/output manipulation functions>.
 
@@ -440,37 +471,45 @@ may or may not be valid. The chunks which you can test for are
 
 =over
 
-=item gAMA
-
-=item sBIT
+=item bKGD
 
 =item cHRM
 
-=item PLTE
-
-=item tRNS
-
-=item bKGD
+=item gAMA
 
 =item hIST
 
-=item pHYs
+=item iCCP
+
+=item IDAT
+
+=item IHDR
+
+=item iTXt
 
 =item oFFs
 
-=item tIME
-
 =item pCAL
 
-=item sRGB
+=item pHYs
 
-=item iCCP
+=item PLTE
 
-=item sPLT
+=item sBIT
 
 =item sCAL
 
-=item IDAT
+=item sPLT
+
+=item sRGB
+
+=item tEXt
+
+=item tIME
+
+=item tRNS
+
+=item zTXt
 
 =back
 
@@ -1007,7 +1046,7 @@ instances of unevaluated arguments, which have all been eliminated
 from this module.
 
 If you are interested in exactly which libpng arguments are omitted,
-you can find each instance L<in the file C<perl-libpng.c> in the top directory of the distribution|http://cpansearch.perl.org/src/BKB/Image-PNG-0.03/perl-libpng.c> in the macro
+you can find each instance L<in the file C<perl-libpng.c> in the top directory of the distribution|http://cpansearch.perl.org/src/BKB/Image-PNG-0.04/perl-libpng.c> in the macro
 C<UNUSED_ZERO_ARG>.
 
 =head2 Function return values are used to return values
@@ -1060,6 +1099,214 @@ program's data area between calls. So this module doesn't deal with
 png_write_chunk, png_write_end, png_write_info, png_write_row, or
 png_write_rows.
 
+=head1 DIAGNOSTICS
+
+The module may produce the following error or warning messages.
+
+=over
+
+
+=item libpng error: %s\n
+
+(F) 
+An error from libpng sent via Perl's warning handler.
+
+
+
+
+=item libpng warning: %s\n
+
+(W) 
+A warning from libpng sent via Perl's warning handler.
+
+
+
+
+=item %s:%d: Call to calloc for %d '%s' failed: out of memory
+
+(F) 
+A request for more memory was refused. The first two parameters are the file name and line number of where this happened.
+
+
+
+
+=item Memory leak detected: there are %d allocated pieces of memory which have not been freed.\n
+
+(W) 
+The module's internal check for memory errors was tripped somehow. This probably indicates a bug in the module.
+
+
+
+
+=item Attempt to destroy an object of unknown type
+
+(F) 
+There was an attempt to free some corrupted memory.
+
+
+
+
+=item According to its compression type, a text chunk in the current PNG file claims to be ITXT but Perl's 'is_utf8_string' says that its encoding is invalid.
+
+(W) 
+
+
+
+=item A language key 'lang_key' member of a 'png_text' structure in the file failed Perl's 'is_utf8_string' test, which says that its encoding is invalid.
+
+(W) 
+
+
+
+=item None of your text chunks was allowed
+
+(W) 
+The user tried to set some text chunks in the image but they were not allowed.
+
+
+
+
+=item Trying to read from a PNG in memory but there is no PNG in memory
+
+(F) 
+Something went wrong trying to read a PNG from a Perl scalar. This probably indicates a bug in the program.
+
+
+
+
+=item Request for too many bytes %d on a scalar of length %d at read position %d.\n
+
+(F) 
+There was an attempt to read some data from a Perl scalar which went beyond the expected end of the scalar in memory.
+
+
+
+
+=item set_IHDR: Bad values for width (%d), height (%d), or bit depth (%d)
+
+(F) 
+The user tried to set a PNG header with unacceptable values, as indicated.
+
+
+
+
+=item set_PLTE: Empty array of colours in set_PLTE
+
+(F) 
+The user tried to set an empty palette of colours.
+
+
+
+
+=item Too many transparencies %d supplied
+
+(F) 
+The user tried to set more than the maximum possible number of transparencies for a paletted image.
+
+
+
+
+=item Image has no height
+
+(F) 
+The height of the image is zero.
+
+
+
+
+=item Image has no rows
+
+(F) 
+The image does not have any rows of image data.
+
+
+
+
+=item Image rows have zero length
+
+(F) 
+The rows of image data have zero length.
+
+
+
+
+=item Image has zero height
+
+(F) 
+The image we are trying to read has zero height.
+
+
+
+
+=item Out of memory allocating %d bytes for image
+
+(F) 
+We were refused the memory we want to read the image into.
+
+
+
+
+=item Row pointers already allocated
+
+(F) 
+There was an attempt to set the rows of an image after they had already been set.
+
+
+
+
+=item Chunk name '%s' has wrong number of characters: %d required
+
+(F) 
+We tried to set an unknown chunk with a name which doesn't have four characters.
+
+
+
+
+=item Number of unknown chunks is zero
+
+(F) 
+The user tried to set an empty list of unknown chunks.
+
+
+
+
+=item Illegal PNG chunk name length, chunk names must be %d characters long
+
+(W) 
+The user's name for a private chunk was not a valid length. In this case the chunk is ignored.
+
+
+
+
+=item Unknown chunk location is zero, which means 'don't write the chunk'
+
+(W) 
+The user supplied a value for "location" of zero, which means "don't write the chunk".
+
+
+
+
+=item undefined chunk name at offset %d in chunk list
+
+(F) 
+
+
+
+=item chunk %i has bad length %d: should be %d in chunk list
+
+(F) 
+
+
+
+=item set_keep_unknown_chunks is not supported in your libpng
+
+(F) 
+libpng was compiled without this option.
+
+
+
+
+=back
 
 =head1 BUGS
 
@@ -1077,15 +1324,13 @@ you will get a memory error.
 =head2 Conditional compilation
 
 It is possible to compile a version of the libpng library without
-support for, for example, text chunks by undefining a macro
-(C<PNG_TEXT_SUPPORTED>). However, this module ignores your libpng
-compilation choices, so it won't compile if you have such a libpng.
-
-I don't know whether many people in practice actually have such a
-libpng without text support or the other optional facilities, but if
-you encounter problems using this Perl module because of your
-conditionally-compiled libpng, then please let me know and I'll
-consider adding that facility to the module.
+support for various things. For example, it's possible to have a
+libpng without support for text chunks by undefining a macro
+C<PNG_TEXT_SUPPORTED>. The module supports some of the conditional
+compilation choices which I've found in practice, but it does not
+support every possible choice. If you encounter problems using this
+Perl module because of a conditionally-compiled libpng, then please
+let me know and I'll consider adding that facility to the module.
 
 =head1 SEE ALSO
 
@@ -1122,6 +1367,70 @@ by actually looking at the libpng source code. These pages are usually
 the first hits on search engines if you search for a function name in
 libpng.
 
+=head2 Other Perl modules on CPAN
+
+=head3 Image::ExifTool
+
+L<Image::ExifTool> is a pure Perl (doesn't require a C compiler)
+solution for accessing the text segments of images. It has extensive
+support for PNG text segments.
+
+=head3 Alien::PNG
+
+L<Alien::PNG> claims to be a way of "building, finding and using PNG
+binaries". It may help in installing libpng. I didn't use it as a
+dependency for this module because it seems not to work in batch mode,
+but stop and prompt the user. I'm interested in hearing feedback from
+users whether this works or not on various platforms.
+
+=head3 Image::PNG::Rewriter
+
+L<Image::PNG::Rewriter> is a utility for unpacking and recompressing
+the IDAT (image data) part of a PNG image. The main purpose seems to
+be to recompress the image data with the module author's other module
+L<Compress::Deflate7>. Unfortunately that only works with Perl
+versions 5.12.
+
+=head3 Image::Pngslimmer
+
+L<Image::Pngslimmer> reduces the size of dynamically created PNG
+images. It's very, very slow at reading PNG data, but seems to work
+OK.
+
+=head3 Image::Info
+
+L<Image::Info> is a module for getting information out of various
+types of images. It has good support for PNG and is written in pure
+Perl (doesn't require a C compiler). As well as basics such as height,
+width, and colour type, it can get text chunks, modification time,
+palette, gamma (gAMA chunk), resolution (pHYs chunk), and significant
+bits (sBIT chunk). At the time of writing (version 1.31) it doesn't
+support other chunks.
+
+
+
+=head1 EXAMPLES
+
+There is a collection of example scripts online at L<http://www.lemoda.net/image-png/>. This currently contains
+
+=over
+
+=item PNG inspector L<http://www.lemoda.net/png/inspect/>
+
+This downloads a PNG file you specify from the internet and prints out
+its contents.
+
+=item PNG quantizer L<http://www.lemoda.net/png/quantize/>
+
+This downloads a PNG file you specify from the internet and quantizes
+it with as many colours as you want (from two to 256) in order to
+reduce its size. The quantized image is uploaded to imgur.com using
+L<WWW::Imgur>.
+
+=back
+
+
+
 =head1 AUTHOR
 
 Ben Bullock, <bkb@cpan.org>
@@ -1161,16 +1470,17 @@ There is a bug tracker at .
 improvement to this module, and who want to share the change with
 other people by adding it to the public version of this module.)
 
-If you want to alter this module, please note very carefully that the
-distributed files are not actually the source code of the module. If
-you wish to obtain the source code of Image::PNG, please access
-http://www.lemoda.net/projects/image-png/repo/ using the "git" source control system.
+If you want to alter this module, note very carefully that the
+distributed files are not actually the source code of the module. The
+source code lives in the "tmpl" directory of the distribution and the
+distribution is created via scripts.
 
 The original files of this distribution make very heavy use of the
 L<Template> Perl module in order to cut down the amount of repetitive
 stuff which needs to be put into various source and documentation
-files. If you plan to alter the C file, you also need a program called
-"cfunctions" which you can download from sourceforge.
+files. If you plan to alter the C file, you may also need a program
+called "cfunctions" which you can download from sourceforge, which
+generates the header file perl-libpng.h from perl-libpng.c.
 
 
 
