@@ -19,10 +19,24 @@ typedef struct perl_libpng {
     png_infop info;
     png_infop end_info;
     enum {perl_png_unknown_obj, perl_png_read_obj, perl_png_write_obj} type;
-    /* Number of rows. */
-    void * row_pointers;
+    /* Allocated memory which holds the rows. */
+    png_bytepp  row_pointers;
+    /* Allocated memory which holds the image data. */
+    png_bytep image_data;
     /* Number of times we have called calloc. */
     int memory_gets;
+
+    /* Jive for reading from a scalar */
+
+    /* Points to the raw data from a Perl scalar. */
+    void * scalar_data;
+    /* Contains what Perl says the length of the data in "image_data"
+       is. */
+    int data_length;
+    /* How much of the data in "image_data" we have read. */
+    int read_position;
+
+
     /* If the following variable is set to a true value, the module
        prints messages about what it is doing. */
     int verbosity : 1;
@@ -72,6 +86,7 @@ typedef perl_libpng_t * Image__PNG__Libpng;
         if (png->verbosity) {                           \
             printf ("%s:%d: ", __FILE__, __LINE__);     \
             printf (x);                                 \
+            printf ("\n");                              \
         }                                               \
     }
 #else
@@ -131,6 +146,7 @@ static void
 perl_png_error_fn (png_structp png_ptr, png_const_charp error_msg)
 {
     perl_libpng_t * png = png_get_error_ptr (png_ptr);
+    /* An error from libpng sent via Perl's warning handler. */
     perl_png_error (png, "libpng error: %s\n", error_msg);
 }
 
@@ -140,6 +156,7 @@ static void
 perl_png_warning_fn (png_structp png_ptr, png_const_charp warning_msg)
 {
     perl_libpng_t * png = png_get_error_ptr (png_ptr);
+    /* A warning from libpng sent via Perl's warning handler. */
     perl_png_warn (png, "libpng warning: %s\n", warning_msg);
 }
 
@@ -166,6 +183,9 @@ perl_png_warning_fn (png_structp png_ptr, png_const_charp warning_msg)
 #define GET_MEMORY(thing, number) {                                     \
         thing = calloc (number, sizeof (*thing));                       \
         if (! thing) {                                                  \
+            /* A request for more memory was refused. The first two
+               parameters are the file name and line number of where
+               this happened. */                \
             perl_png_error (png, "%s:%d: Call to calloc for "           \
                             "%d '%s' failed: out of memory",            \
                             __FILE__, __LINE__, number, #thing);        \
@@ -224,8 +244,16 @@ perl_png_create_read_struct ()
 
 static void free_png (perl_libpng_t * png)
 {
-    MESSAGE ("Freeing PNG memory.\n");
+    MESSAGE ("Freeing PNG memory.");
+    if (png->row_pointers) {
+        PERL_PNG_FREE (png->row_pointers);
+    }
+    if (png->image_data) {
+        PERL_PNG_FREE (png->image_data);
+    }
     if (png->memory_gets != 1) {
+        /* The module's internal check for memory errors was tripped
+           somehow. This probably indicates a bug in the module. */
         perl_png_warn (png, "Memory leak detected: there are %d "
                        "allocated pieces of memory which have not "
                        "been freed.\n", png->memory_gets - 1);
@@ -237,9 +265,6 @@ void
 perl_png_destroy_write_struct (perl_libpng_t * png)
 {
     png_destroy_write_struct (& png->png, & png->info);
-    if (png->row_pointers) {
-        PERL_PNG_FREE (png->row_pointers);
-    }
     free_png (png);
 }
 
@@ -266,6 +291,7 @@ perl_png_destroy (perl_libpng_t * png)
         perl_png_destroy_read_struct (png);
     }
     else {
+        /* There was an attempt to free some corrupted memory. */
         perl_png_error (png, "Attempt to destroy an object of unknown type");
     }
 }
@@ -441,7 +467,7 @@ perl_png_get_text (perl_libpng_t * png)
         SV * text_ref;
         AV * text_chunks;
 
-        MESSAGE ("Got some text:\n");
+        MESSAGE ("Got some text:");
         text_chunks = newAV ();
         for (i = 0; i < num_text; i++) {
             HV * hash;
@@ -456,7 +482,7 @@ perl_png_get_text (perl_libpng_t * png)
         text_ref = newRV_inc ((SV *) text_chunks);
         return text_ref;
     } else {
-        MESSAGE ("There is no text:\n");
+        MESSAGE ("There is no text:");
         UNDEF;
     }
 }
@@ -467,18 +493,19 @@ static int
 perl_png_set_text_from_hash (perl_libpng_t * png, png_text * text_out, HV * chunk)
 {
     int compression;
-    char * keyword;
-    int keyword_length;
+    char * key;
+    int key_length;
+#ifdef PNG_iTXt_SUPPORTED
     char * language_tag;
     int language_tag_length;
-    char * translated_keyword;
-    int translated_keyword_length;
+    char * translated_key;
+#endif /* PNG_iTXt_SUPPORTED */
     int is_itxt = 0;
     char * text;
     int text_length;
     int ok = 1;
 
-    MESSAGE ("Putting it into something.\n");
+    MESSAGE ("Putting it into something.");
     /* Check the compression field of the chunk */
 #define FETCH_IV(field) {                                       \
         SV * field_sv;                                          \
@@ -527,28 +554,28 @@ perl_png_set_text_from_hash (perl_libpng_t * png, png_text * text_out, HV * chun
         }                                                       \
     }
 
-    MESSAGE ("Getting keyword.\n");
-    FETCH_PV (keyword);
-    if (keyword_length < 1 || keyword_length > 79) {
-        /* Keyword is too long or empty */
-        MESSAGE ("Bad length of keyword.\n");
+    MESSAGE ("Getting key.");
+    FETCH_PV (key);
+    if (key_length < 1 || key_length > 79) {
+        /* Key is too long or empty */
+        MESSAGE ("Bad length of key.");
 
         ok = 0;
         return 0;
     }
-    MESSAGE ("Getting text.\n");
+    MESSAGE ("Getting text.");
     FETCH_PV (text);
     if (ok) {
-        MESSAGE ("Copying.\n");
+        MESSAGE ("Copying.");
         text_out->compression = compression;
-        text_out->key = keyword;
+        text_out->key = key;
         text_out->text = text;
         text_out->text_length = text_length;
 #ifdef PNG_iTXt_SUPPORTED
         if (is_itxt) {
             FETCH_PV (language_tag);
             text_out->lang = language_tag;
-            text_out->lang_key = translated_keyword;
+            text_out->lang_key = translated_key;
         }
 #endif
     }
@@ -587,20 +614,20 @@ perl_png_set_text (perl_libpng_t * png, AV * text_chunks)
         chunk_pointer = av_fetch (text_chunks, i, 0);
         if (! chunk_pointer) {
             /* Complain */
-            MESSAGE ("Chunk pointer null.\n");
+            MESSAGE ("Chunk pointer null.");
             continue;
         }
         if (SvROK (* chunk_pointer) && 
             SvTYPE (SvRV (* chunk_pointer)) == SVt_PVHV) {
-            MESSAGE ("Looks like a hash.\n");
+            MESSAGE ("Looks like a hash.");
             ok = perl_png_set_text_from_hash (png, & png_texts[num_ok],
                                               (HV *) SvRV (* chunk_pointer));
             if (ok) {
-                MESSAGE ("This chunk is OK.\n");
+                MESSAGE ("This chunk is OK.");
                 num_ok++;
             }
             else {
-                MESSAGE ("The chunk is not OK.\n");
+                MESSAGE ("The chunk is not OK.");
             }
         }
     }
@@ -610,6 +637,8 @@ perl_png_set_text (perl_libpng_t * png, AV * text_chunks)
         png_set_text (pngi, png_texts, num_ok);
     }
     else {
+        /* The user tried to set some text chunks in the image but
+           they were not allowed. */
         perl_png_warn (png, "None of your text chunks was allowed");
     }
     PERL_PNG_FREE (png_texts);
@@ -664,7 +693,7 @@ void perl_png_timep_to_hash (const png_timep mod_time, HV * time_hash)
    associative array. */
 
 SV *
-perl_png_get_time (perl_libpng_t * png)
+perl_png_get_tIME (perl_libpng_t * png)
 {
     png_timep mod_time = 0;
     int status;
@@ -684,7 +713,7 @@ perl_png_get_time (perl_libpng_t * png)
 /* Set the time in the PNG from "time". */
 
 void
-perl_png_set_time (perl_libpng_t * png, SV * input_time)
+perl_png_set_tIME (perl_libpng_t * png, SV * input_time)
 {
     png_time mod_time = {0};
     SV * ref;
@@ -692,7 +721,7 @@ perl_png_set_time (perl_libpng_t * png, SV * input_time)
     ref = SvRV(input_time);
     if (ref && SvTYPE (ref) == SVt_PVHV) {
         HV * time_hash = (HV *) SvRV (input_time);
-        MESSAGE ("Setting time from a hash.\n");
+        MESSAGE ("Setting time from a hash.");
 #define SET_TIME(field) {                                               \
             SV ** field_sv_ptr = hv_fetch (time_hash, #field,           \
                                            strlen (#field), 0);         \
@@ -718,7 +747,7 @@ perl_png_set_time (perl_libpng_t * png, SV * input_time)
            now, according to the system clock. */
         time_t now;
 
-        MESSAGE ("The modification time doesn't look OK so I am going to set the modification time to the time now instead.\n");
+        MESSAGE ("The modification time doesn't look OK so I am going to set the modification time to the time now instead.");
         now = time (0);
         png_convert_from_time_t (& mod_time, now);
     }
@@ -729,7 +758,7 @@ int
 perl_png_sig_cmp (SV * png_header, int start, int num_to_check)
 {
     unsigned char * header;
-    unsigned int length;
+    int length;
     int ret_val;
     header = (unsigned char *) SvPV (png_header, length);
     ret_val = png_sig_cmp (header, start, num_to_check);
@@ -752,33 +781,70 @@ typedef struct {
     const char * data; 
     int read_position;
     unsigned int length;
+    perl_libpng_t * png;
 }
 scalar_as_image_t;
 
 /* Read some bytes from a Perl scalar into a png->png as requested. */
 
 static void
-perl_png_scalar_read (png_structp png,
+perl_png_scalar_read (png_structp pngstruct,
                       png_bytep out_bytes,
                       png_size_t byte_count_to_read)
 {
-    scalar_as_image_t * si;
-    const char * read_point;
+    const unsigned char * read_point;
+    perl_libpng_t * png;
 
-    si = png_get_io_ptr (png);
-#if 0
-    fprintf (stderr, "Reading %d bytes from image at position %d.\n",
-             byte_count_to_read, si->read_position);
-#endif
-    if (si->read_position + byte_count_to_read > si->length) {
-        fprintf (stderr, "Request for too many bytes %d on an image "
-                 "of length %d at read position %d.\n",
-                 byte_count_to_read, si->length, si->read_position);
+    png = png_get_io_ptr (pngstruct);
+    if (! png->scalar_data) {
+        /* Something went wrong trying to read a PNG from a Perl
+           scalar. This probably indicates a bug in the program. */
+        perl_png_error (png, "Trying to read from a PNG in memory but there is no PNG in memory");
+    }
+
+    MESSAGE ("Length of data is %d. "
+             "Read position is %d. "
+             "Length to read is %d. ",
+             png->data_length, png->read_position,
+             byte_count_to_read);
+    /*
+    fprintf (stderr, "Reading %d bytes from image length %d at position %d\n",
+             byte_count_to_read, png->data_length, png->read_position);
+    */
+    if (png->read_position + byte_count_to_read > png->data_length) {
+        /* There was an attempt to read some data from a Perl scalar
+           which went beyond the expected end of the scalar in
+           memory. */
+        perl_png_error (png, "Request for too many bytes %d on a scalar "
+                        "of length %d at read position %d.\n",
+                        byte_count_to_read, png->data_length,
+                        png->read_position);
         return;
     }
-    read_point = si->data + si->read_position;
+    read_point = png->scalar_data + png->read_position;
     memcpy (out_bytes, read_point, byte_count_to_read);
-    si->read_position += byte_count_to_read;
+    png->read_position += byte_count_to_read;
+}
+
+void
+perl_png_scalar_as_input (perl_libpng_t * png,
+                          SV * image_data,
+                          int transforms)
+{
+    MESSAGE ("Setting input from a scalar");
+
+    /* We don't need the following anywhere. However we probably
+       should keep track of where the data comes from. */
+    png->scalar_data = SvPV (image_data, png->data_length);
+
+    MESSAGE ("Length of data is %d. "
+            "Read position is %d.",
+            png->data_length, png->read_position);
+
+
+    MESSAGE ("Length of the scalar data is %d", png->data_length);
+    /* Set the reader for png->png to our function. */
+    png_set_read_fn (png->png, png, perl_png_scalar_read);
 }
 
 /* Read a PNG from a Perl scalar "image_data". */
@@ -788,16 +854,7 @@ perl_png_read_from_scalar (perl_libpng_t * png,
                            SV * image_data,
                            int transforms)
 {
-    scalar_as_image_t si = {0};
-
-    /* We don't need the following anywhere. However we probably
-       should keep track of where the data comes from. */
-    si.png_image = image_data;
-    si.data = SvPV (si.png_image, si.length);
-    /* Check it is a valid PNG here using png_sig_cmp. */
-
-    /* Set the reader for png->png to our function. */
-    png_set_read_fn (png->png, & si, perl_png_scalar_read);
+    perl_png_scalar_as_input (png, image_data, transforms);
     png_read_png (png->png, png->info, transforms, UNUSED_ZERO_ARG);
 }
 
@@ -825,7 +882,7 @@ perl_png_write_to_scalar (perl_libpng_t * png, int transforms)
     SV * image_data;
 
     GET_MEMORY (si, 1);
-    MESSAGE ("Setting up the image.\n");
+    MESSAGE ("Setting up the image.");
     /* Set the writer for png->png to our function. */
     png_set_write_fn (png->png, si, perl_png_scalar_write,
                       0 /* No flush function */);
@@ -838,7 +895,7 @@ perl_png_write_to_scalar (perl_libpng_t * png, int transforms)
 void
 perl_png_write_png (perl_libpng_t * png, int transforms)
 {
-    MESSAGE ("Trying to write your crap.\n");
+    MESSAGE ("Trying to write your crap.");
     png_write_png (pngi, transforms, UNUSED_ZERO_ARG);
 }
 
@@ -912,7 +969,9 @@ perl_png_set_IHDR (perl_libpng_t * png, HV * IHDR)
     FETCH (color_type);
     FETCH (interlace_method);
     if (width == 0 || height == 0 || bit_depth == 0) {
-        perl_png_error (png, "Bad values for width (%d), height (%d), or bit depth (%d)",
+        /* The user tried to set a PNG header with unacceptable values,
+           as indicated. */
+        perl_png_error (png, "set_IHDR: Bad values for width (%d), height (%d), or bit depth (%d)",
                         width, height, bit_depth);
         return;
     }
@@ -1027,7 +1086,8 @@ perl_png_set_PLTE (perl_libpng_t * png, AV * perl_colors)
 
     n_colors = av_len (perl_colors) + 1;
     if (n_colors == 0) {
-        perl_png_error (png, "Empty array of colours in set_PLTE");
+        /* The user tried to set an empty palette of colours. */
+        perl_png_error (png, "set_PLTE: Empty array of colours in set_PLTE");
     }
     MESSAGE ("There are %d colours in the palette.\n", n_colors);
 
@@ -1057,6 +1117,13 @@ perl_png_set_PLTE (perl_libpng_t * png, AV * perl_colors)
     PERL_PNG_FREE (colors);
 }
 
+/* Set the palette directly, using a pointer. */
+
+void
+perl_png_set_PLTE_pointer (perl_libpng_t * png, png_colorp colors, int n_colors){
+    png_set_PLTE (pngi, colors, n_colors);
+}
+
 /* Create a hash containing the colour values of a pointer to a
    png_color_16 structure. */
 
@@ -1096,6 +1163,10 @@ SV * perl_png_get_bKGD (perl_libpng_t * png)
     UNDEF;
 }
 
+void perl_png_set_bKGD (perl_libpng_t * png, HV * bKGD)
+{
+}
+
 SV * perl_png_get_cHRM (perl_libpng_t * png)
 {
     if (VALID (cHRM)) {
@@ -1106,6 +1177,99 @@ SV * perl_png_get_cHRM (perl_libpng_t * png)
     UNDEF;
 }
 
+void perl_png_set_cHRM (perl_libpng_t * png, HV * cHRM)
+{
+}
+
+SV * perl_png_get_pCAL (perl_libpng_t * png)
+{
+    if (VALID (pCAL)) {
+        HV * ice;
+        ice = newHV ();
+        return newRV_inc ((SV *) ice);
+    }
+    UNDEF;
+}
+
+void perl_png_set_pCAL (perl_libpng_t * png, HV * pCAL)
+{
+}
+
+SV * perl_png_get_sPLT (perl_libpng_t * png)
+{
+    if (VALID (sPLT)) {
+        HV * ice;
+        ice = newHV ();
+        return newRV_inc ((SV *) ice);
+    }
+    UNDEF;
+}
+
+void perl_png_set_sPLT (perl_libpng_t * png, HV * sPLT)
+{
+}
+
+SV * perl_png_get_gAMA (perl_libpng_t * png)
+{
+    if (VALID (gAMA)) {
+        HV * ice;
+        ice = newHV ();
+        return newRV_inc ((SV *) ice);
+    }
+    UNDEF;
+}
+
+void perl_png_set_gAMA (perl_libpng_t * png, HV * gAMA)
+{
+}
+
+SV * perl_png_get_iCCP (perl_libpng_t * png)
+{
+    if (VALID (iCCP)) {
+        HV * ice;
+        ice = newHV ();
+        return newRV_inc ((SV *) ice);
+    }
+    UNDEF;
+}
+
+void perl_png_set_iCCP (perl_libpng_t * png, HV * iCCP)
+{
+}
+
+SV * perl_png_get_tRNS (perl_libpng_t * png)
+{
+    if (VALID (tRNS)) {
+        HV * ice;
+        ice = newHV ();
+        return newRV_inc ((SV *) ice);
+    }
+    UNDEF;
+}
+
+void perl_png_set_tRNS (perl_libpng_t * png, HV * tRNS)
+{
+}
+
+void perl_png_set_tRNS_pointer (perl_libpng_t * png, png_bytep trans, int num_trans)
+{
+    png_set_tRNS (pngi, trans, num_trans, 0);
+}
+
+SV * perl_png_get_sCAL (perl_libpng_t * png)
+{
+    if (VALID (sCAL)) {
+        HV * ice;
+        ice = newHV ();
+        return newRV_inc ((SV *) ice);
+    }
+    UNDEF;
+}
+
+void perl_png_set_sCAL (perl_libpng_t * png, HV * sCAL)
+{
+}
+
 SV * perl_png_get_sBIT (perl_libpng_t * png)
 {
     if (VALID (sBIT)) {
@@ -1116,6 +1280,10 @@ SV * perl_png_get_sBIT (perl_libpng_t * png)
     UNDEF;
 }
 
+void perl_png_set_sBIT (perl_libpng_t * png, HV * sBIT)
+{
+}
+
 SV * perl_png_get_oFFs (perl_libpng_t * png)
 {
     if (VALID (oFFs)) {
@@ -1123,10 +1291,14 @@ SV * perl_png_get_oFFs (perl_libpng_t * png)
         
         offset = newHV ();
         return newRV_inc ((SV *) offset);
-     }
-     UNDEF;
- }
+    }
+    UNDEF;
+}
  
+void perl_png_set_oFFs (perl_libpng_t * png, HV * oFFs)
+{
+}
+
 SV * perl_png_get_pHYs (perl_libpng_t * png)
 {
     if (VALID (pHYs)) {
@@ -1137,6 +1309,42 @@ SV * perl_png_get_pHYs (perl_libpng_t * png)
     UNDEF;
 }
 
+void perl_png_set_pHYs (perl_libpng_t * png, HV * pHYs)
+{
+}
+
+/* Get the transparency information for a paletted image. */
+
+SV * perl_png_get_tRNS_palette (perl_libpng_t * png)
+{
+    if (VALID (tRNS) && VALID (PLTE)) {
+        AV * perl_trans;
+        png_bytep png_trans;
+        int num_trans;
+        SV ** perl_trans_sv;
+        int i;
+
+        png_get_tRNS (pngi, & png_trans, & num_trans, 0);
+
+        if (num_trans == 0) {
+            UNDEF;
+        }
+        if (num_trans > 0x100) {
+            /* The user tried to set more than the maximum possible
+               number of transparencies for a paletted image. */
+            perl_png_error (png, "Too many transparencies %d supplied",
+                            num_trans);
+        }
+        perl_trans = newAV ();
+        for (i = 0; i < num_trans; i++) {
+            SV * trans_i = newSViv (png_trans[i]);
+            av_push (perl_trans, trans_i);
+        }
+        return newRV_inc ((SV *) perl_trans);
+    }
+    UNDEF;
+}
+
 int perl_png_get_sRGB (perl_libpng_t * png)
 {
     /* I'm not sure what to return if there is no valid sRGB value. */
@@ -1149,6 +1357,10 @@ int perl_png_get_sRGB (perl_libpng_t * png)
     return intent;
 }
 
+void perl_png_set_sRGB (perl_libpng_t * png, int sRGB)
+{
+}
+
 SV * perl_png_get_valid (perl_libpng_t * png)
 {
     HV * perl_valid;
@@ -1159,22 +1371,7 @@ SV * perl_png_get_valid (perl_libpng_t * png)
 #define V(x) \
     hv_store (perl_valid, #x, strlen (#x), newSViv (valid & PNG_INFO_ ## x), 0)
 
-    V(gAMA);
-    V(sBIT);
-    V(cHRM);
-    V(PLTE);
-    V(tRNS);
-    V(bKGD);
-    V(hIST);
-    V(pHYs);
-    V(oFFs);
-    V(tIME);
-    V(pCAL);
-    V(sRGB);
-    V(iCCP);
-    V(sPLT);
-    V(sCAL);
-    V(IDAT);
+V(bKGD);V(cHRM);V(gAMA);V(hIST);V(iCCP);V(IDAT);V(oFFs);V(pCAL);V(pHYs);V(PLTE);V(sBIT);V(sCAL);V(sPLT);V(sRGB);V(tIME);V(tRNS);
 
 #undef V
 
@@ -1203,6 +1400,7 @@ perl_png_get_rows (perl_libpng_t * png)
 
     height = png_get_image_height (pngi);
     if (height == 0) {
+        /* The height of the image is zero. */
         perl_png_error (png, "Image has no height");
     }
     else {
@@ -1210,13 +1408,15 @@ perl_png_get_rows (perl_libpng_t * png)
     }
     rows = png_get_rows (pngi);
     if (rows == 0) {
+        /* The image does not have any rows of image data. */
         perl_png_error (png, "Image has no rows");
     }
     else {
-        MESSAGE ("Image has some rows\n");
+        MESSAGE ("Image has some rows");
     }
     rowbytes = png_get_rowbytes (pngi);
     if (rowbytes == 0) {
+        /* The rows of image data have zero length. */
         perl_png_error (png, "Image rows have zero length");
     }
     else {
@@ -1236,6 +1436,61 @@ perl_png_get_rows (perl_libpng_t * png)
     return newRV_inc ((SV *) perl_rows);
 }
 
+/* Read the image data into allocated memory */
+
+void
+perl_png_read_image (perl_libpng_t * png)
+{
+    int n_rows;
+    int rowbytes;
+    int i;
+
+    n_rows = png_get_image_height (pngi);
+    rowbytes = png_get_rowbytes (pngi);
+    if (! n_rows) {
+        /* The image we are trying to read has zero height. */
+        perl_png_error (png, "Image has zero height");
+    }
+    GET_MEMORY (png->row_pointers, n_rows);
+    png->image_data = malloc (rowbytes * n_rows);
+    if (! png->image_data) {
+        /* We were refused the memory we want to read the image into. */
+        perl_png_error (png, "Out of memory allocating %d bytes for image",
+                        rowbytes * n_rows);
+    }
+    png->memory_gets++;
+    for (i = 0; i < n_rows; i++) {
+        png->row_pointers[i] = png->image_data + rowbytes * i;
+    }
+    png_read_image (png->png, png->row_pointers);
+    /* Set the row_pointers pointers to point into the allocated
+       memory. */
+}
+
+/* Get the row pointers directly. */
+
+void *
+perl_png_get_row_pointers (perl_libpng_t * png)
+{
+    png_bytepp rows;
+    if (png->row_pointers) {
+        rows = png->row_pointers;
+    }
+    else {
+        rows = png_get_rows (pngi);
+    }
+    return rows;
+}
+
+/* Set the row pointers directly from a pointer. */
+
+void
+perl_png_set_row_pointers (perl_libpng_t * png, void * row_pointers)
+{
+    png_set_rows (pngi, row_pointers);
+    png->row_pointers = 0;
+}
+
 /* Set the rows of the image to "rows". */
 
 void perl_png_set_rows (perl_libpng_t * png, AV * rows)
@@ -1245,6 +1500,8 @@ void perl_png_set_rows (perl_libpng_t * png, AV * rows)
     int n_rows;
 
     if (png->row_pointers) {
+        /* There was an attempt to set the rows of an image after they
+           had already been set. */
         perl_png_error (png, "Row pointers already allocated");
     }
     /* Need to check that this is the same as the height of the image. */
@@ -1281,7 +1538,7 @@ void perl_png_set_rows (perl_libpng_t * png, AV * rows)
 void perl_png_set_verbosity (perl_libpng_t * png, int verbosity)
 {
     png->verbosity = verbosity;
-    MESSAGE ("You have asked me to print messages saying what I'm doing.\n");
+    MESSAGE ("You have asked me to print messages saying what I'm doing.");
 }
 
 /* What should we do when there is an undefined variable, print an
@@ -1347,7 +1604,9 @@ SV * perl_png_get_unknown_chunks (perl_libpng_t * png)
 
             png_chunk = unknown_chunks + i;
             perl_chunk = newHV ();
-            if (strlen (png_chunk->name) != PERL_PNG_CHUNK_NAME_LENGTH) {
+            if (strlen ((const char *) png_chunk->name) != PERL_PNG_CHUNK_NAME_LENGTH) {
+                /* We tried to set an unknown chunk with a name which
+                   doesn't have four characters. */
                 perl_png_error (png, "Chunk name '%s' has wrong number of "
                                 "characters: %d required",
                                 png_chunk->name,
@@ -1383,11 +1642,12 @@ void perl_png_set_unknown_chunks (perl_libpng_t * png, AV * chunk_list)
     n_chunks = av_len (chunk_list) + 1;
    
     if (n_chunks == 0) {
+        /* The user tried to set an empty list of unknown chunks. */
         perl_png_error (png, "Number of unknown chunks is zero");
     }
     GET_MEMORY (unknown_chunks, n_chunks);
     n_ok_chunks = 0;
-    MESSAGE ("OK.\n");
+    MESSAGE ("OK.");
     for (i = 0; i < n_chunks; i++) {
         HV * perl_chunk;
         png_unknown_chunk * png_chunk;
@@ -1402,6 +1662,8 @@ void perl_png_set_unknown_chunks (perl_libpng_t * png, AV * chunk_list)
 
         HASH_FETCH_PV (perl_chunk, name);
         if (name_length != PERL_PNG_CHUNK_NAME_LENGTH) {
+            /* The user's name for a private chunk was not a valid
+               length. In this case the chunk is ignored. */
             perl_png_warn (png, "Illegal PNG chunk name length, "
                            "chunk names must be %d characters long",
                            PERL_PNG_CHUNK_NAME_LENGTH);
@@ -1421,7 +1683,8 @@ void perl_png_set_unknown_chunks (perl_libpng_t * png, AV * chunk_list)
 
         HASH_FETCH_IV (perl_chunk, location);
         if (location == 0) {
-            /* warn this means "don't write the chunk". */
+            /* The user supplied a value for "location" of zero, which
+               means "don't write the chunk". */
             perl_png_warn (png, "Unknown chunk location is zero, which means 'don't write the chunk'");
         }
         png_chunk->location = location;
@@ -1443,6 +1706,10 @@ int perl_png_supports (const char * what)
         return 0;
 #endif /* iTXt */
     }
+    /* The user asked whether something was supported, but we don't
+       know what that thing is. */
+    perl_png_warn (0, "Unknown string '%s'", what);
+    return 0;
 }
 
 void
@@ -1505,15 +1772,39 @@ perl_png_set_keep_unknown_chunks (perl_libpng_t * png, int keep,
         png->memory_gets--;
     }
     else {
-        MESSAGE ("There is no valid chunk list.\n");
+        MESSAGE ("There is no valid chunk list.");
     empty_chunk_list:
         png_set_keep_unknown_chunks (png->png, keep, 0, 0);
     }
 #else
-    perl_png_error ("This function is not supported in your libpng");
+    /* libpng was compiled without this option. */
+    perl_png_error (png, "set_keep_unknown_chunks is not supported in your libpng");
 #endif
 }
-                                  
+
+void
+perl_png_set_expand (perl_libpng_t * png)
+{
+    png_set_expand (png->png);
+}
+
+void
+perl_png_set_gray_to_rgb (perl_libpng_t * png)
+{
+    png_set_gray_to_rgb(png->png);
+}
+
+void
+perl_png_set_filler (perl_libpng_t * png, int filler, int flags)
+{
+    png_set_filler (png->png, filler, flags);
+}
+
+void
+perl_png_set_strip_16 (perl_libpng_t * png)
+{
+    png_set_strip_16 (png->png);
+}
 
 /*
    Local Variables:
